/*
 * Algebraic manipulator complex number routines.
 *
 * Copyright (c) 1996 George Gesslein II.
 */

#include "am.h"
#include "externs.h"

/*
 * Multiply two complex numbers and return result.
 */
complexs
cmult(a, b)
complexs	a;
complexs	b;
{
	complexs	r;

	r.re = a.re * b.re - a.im * b.im;
	r.im = a.re * b.im + a.im * b.re;
	return(r);
}

/*
 * Take the natural logarithm of a complex number and return result.
 */
complexs
clog(a)
complexs	a;
{
	complexs	r;

	r.re = log(a.re * a.re + a.im * a.im) / 2.0;
	r.im = atan2(a.im, a.re);
	return(r);
}

/*
 * Raise the natural number (e) to the power of a complex number
 * and return result.
 */
complexs
cexp(a)
complexs	a;
{
	complexs	r;
	double		m;

	m = exp(a.re);
	r.re = m * cos(a.im);
	r.im = m * sin(a.im);
	return(r);
}

/*
 * Raise a complex number to the power of a real number and return result.
 */
complexs
cpow(a, b)
complexs	a;
double		b;
{
	complexs	r;

	r = clog(a);
	r.re *= b;
	r.im *= b;
	r = cexp(r);
	return(r);
}

int
roots(cp)
char	*cp;
{
	complexs	c, c2, check;
	double		d, k;
	double		root;
	double		radius, theta;
	double		radius_root;
	char		*cp1;

	if (extra_garbage(cp))
		return false;
	for (;;) {
		strcpy(prompt_str, _("Enter root (positive integer): "));
		if ((cp1 = getstring((char *) &scratch[0], n_tokens * sizeof(token_type))) == NULL)
			return false;
		if (*cp1 == '\0' || *cp1 == '\n') {
			return false;
		}
		root = atof(cp1);
		if (root <= 0.0 || fmod(root, 1.0) != 0.0) {
			printf(_("Root must be positive integer.\n"));
			continue;
		}
		break;
	}
	strcpy(prompt_str, _("Enter real part: "));
	if ((cp1 = getstring((char *) &scratch[0], n_tokens * sizeof(token_type))) == NULL)
		return false;
	c.re = atof(cp1);
	strcpy(prompt_str, _("Enter imaginary part: "));
	if ((cp1 = getstring((char *) &scratch[0], n_tokens * sizeof(token_type))) == NULL)
		return false;
	c.im = atof(cp1);
	if (c.re == 0.0 && c.im == 0.0)
		return false;
/* convert to polar coordinates */
	radius = sqrt(c.re * c.re + c.im * c.im);
	theta = atan2(c.im, c.re);
	radius_root = pow(radius, 1.0 / root);
	printf(_("The %.14lg roots of (%.14lg + %.14lg*i#)^(1/%.14lg) are:\n\n"), root, c.re, c.im, root);
	for (k = 0.0; k < root; k += 1.0) {
		c2.re = radius_root * cos((theta + 2.0 * k * PI) / root);
		c2.im = radius_root * sin((theta + 2.0 * k * PI) / root);
		fprintf(gfp, "%.14lg + %.14lg*i#\n", c2.re, c2.im);
		check = c2;
		for (d = 1.0; d < root; d += 1.0) {
			check = cmult(check, c2);
		}
		printf(_("Inverse Check: %.14lg + %.14lg*i#\n\n"), check.re, check.im);
	}
	return true;
}

/*
 * Simplify roots of complex numbers.
 * Only gives one root, so this is not perfect.
 *
 * Returns true if expression was modified.
 */
int
complex_root_simp(equation, np)
token_type	*equation;
int		*np;
{
	int		i, j;
	int		imag_cnt, plus_cnt, times_cnt;
	int		level, level2;
	double		power;
	complexs	c;

	for (i = 1; i < *np; i += 2) {
		if (equation[i].token.operatr != POWER)
			continue;
		level = equation[i].level;
		if (equation[i+1].level != level
		    || equation[i+1].kind != CONSTANT)
			continue;
		power = equation[i+1].token.constant;
		c.re = 0.0;
		c.im = 1.0;
		imag_cnt = 0;
		plus_cnt = 0;
		times_cnt = 0;
		for (j = i - 1; j >= 0 && equation[j].level >= level; j--) {
			switch (equation[j].kind) {
			case VARIABLE:
				if (equation[j].token.variable != IMAGINARY)
					goto end_loop;
				imag_cnt++;
				break;
			case OPERATOR:
				switch (equation[j].token.operatr) {
				case TIMES:
					times_cnt++;
					level2 = equation[j].level;
					if (equation[j-1].level != level2
					    || equation[j+1].level != level2)
						goto end_loop;
					if (equation[j-1].kind == VARIABLE
					    && equation[j-1].token.variable == IMAGINARY) {
						if (equation[j+1].kind != CONSTANT)
							goto end_loop;
						c.im = equation[j+1].token.constant;
						continue;
					}
					if (equation[j+1].kind == VARIABLE
					    && equation[j+1].token.variable == IMAGINARY) {
						if (equation[j-1].kind != CONSTANT)
							goto end_loop;
						c.im = equation[j-1].token.constant;
						continue;
					}
					goto end_loop;
				case MINUS:
					if (times_cnt || imag_cnt) {
						c.im = -c.im;
					}
				case PLUS:
					if (++plus_cnt > 1)
						goto end_loop;
					level2 = equation[j].level;
					if (equation[j-1].level == level2
					    && equation[j-1].kind == CONSTANT) {
						c.re = equation[j-1].token.constant;
						continue;
					}
					if (equation[j+1].level == level2
					    && equation[j+1].kind == CONSTANT) {
						c.re = equation[j+1].token.constant;
						if (equation[j].token.operatr == MINUS)
							c.re = -c.re;
						continue;
					}
				default:
					goto end_loop;
				}
			case CONSTANT:
				break;
			default:
				goto end_loop;
			}
		}
		if (imag_cnt != 1)
			goto end_loop;
		j++;
		i += 2;
#if	!SILENT
		printf(_("Warning: Complex number root taken.\n"));
#endif
		c = cpow(c, power);
		if (*np + 5 - (i - j) > n_tokens) {
			error_huge();
		}
		blt(&equation[j+5], &equation[i], (*np - i) * sizeof(token_type));
		*np += 5 - (i - j);
		equation[j].level = level;
		equation[j].kind = CONSTANT;
		equation[j].token.constant = c.re;
		j++;
		equation[j].level = level;
		equation[j].kind = OPERATOR;
		equation[j].token.operatr = PLUS;
		j++;
		equation[j].level = level + 1;
		equation[j].kind = CONSTANT;
		equation[j].token.constant = c.im;
		j++;
		equation[j].level = level + 1;
		equation[j].kind = OPERATOR;
		equation[j].token.operatr = TIMES;
		j++;
		equation[j].level = level + 1;
		equation[j].kind = VARIABLE;
		equation[j].token.variable = IMAGINARY;
		return true;
end_loop:
		;
	}
	return false;
}
